/*
* Copyright (C) 2012 Chris Neasbitt
* Author: Chris Neasbitt
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package edu.uga.cs.adblock;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
//import org.apache.commons.logging.Log;
//import org.apache.commons.logging.LogFactory;
import org.json.JSONException;

import com.google.common.net.InternetDomainName;

import edu.uga.cs.json.JSONReader;

public class AdBlockRuleMatcher {

	private static transient Log log = LogFactory.getLog(AdBlockRuleMatcher.class);
	
	private final List<AdBlockRule> rules;
	private final Map<String, Boolean> isBlockedIndex = 
			new HashMap<String, Boolean>();
	
	//private static final transient Log log = LogFactory.getLog(AdBlockRuleMatcher.class);

	public AdBlockRuleMatcher() throws Exception {
		rules = parseDefaultRules();
	}
	
	public AdBlockRuleMatcher(String parsedRulesFile) throws IOException,
			JSONException {
		rules = loadRules(parsedRulesFile);
	}

	public AdBlockRuleMatcher(List<AdBlockRule> rules) {
		this.rules = rules;
	}
	
	private List<AdBlockRule> parseDefaultRules() throws Exception{
		InputStream list = AdBlockRuleMatcher.class
				.getClassLoader().getResourceAsStream("easylist.txt");
		return AdBlockRuleParser.parseRules(list);
	}

	private List<AdBlockRule> loadRules(String parsedRulesFile) throws IOException,
			JSONException {
		JSONReader<AdBlockRule> reader = new JSONReader<AdBlockRule>();
		return reader.read(new File(parsedRulesFile), AdBlockRule.class);
	}

	public boolean isBlocked2LD(String url){
		try {
			URL urlObj = new URL(url);
			InternetDomainName dname = InternetDomainName.from(urlObj.getHost());
			String shortUrl = urlObj.getProtocol() + "://" + dname.topPrivateDomain().name() + "/";
			return isBlocked(shortUrl);
		} catch(Exception e){
			if(log.isErrorEnabled()){
				log.error("Error matching url.", e);
			}
			return false;
		}
	}
	
	public boolean isBlockedFull2LD(String url){
		try{
			URL urlObj = new URL(url);
			InternetDomainName dname = InternetDomainName.from(urlObj.getHost());
			String pathStr = urlObj.getPath();
			String queryStr = urlObj.getQuery();
			String urlEnd = "";
			if(pathStr != null){
				urlEnd += pathStr;
			}
			if(queryStr != null){
				urlEnd += "?" + queryStr;
			}
			String twoLDUrl = urlObj.getProtocol() + "://" + dname.topPrivateDomain().name() +
					urlEnd;
			
			return isBlocked(twoLDUrl);
		} catch(Exception e){
			if(log.isErrorEnabled()){
				log.error("Error matching url.", e);
			}
			return false;
		}
	}
	
	public boolean isBlocked(String url){
		try{
			if(isBlockedIndex.containsKey(url)){
				return isBlockedIndex.get(url);
			}
			List<AdBlockRule> rules = this.getMatchedRules(url, true);
			filterExceptionRules(rules);
			boolean retval = rules.size() > 0;
			isBlockedIndex.put(url, retval);
			return retval;
		} catch (Exception e){
			if(log.isErrorEnabled()){
				log.error("Error matching url.", e);
			}
			return false;
		}
	}

	private void filterExceptionRules(List<AdBlockRule> rules) {
		List<AdBlockRule> remove = new ArrayList<AdBlockRule>();
		for (AdBlockRule rule : rules) {
			if (rule.isException()) {
				remove.add(rule);
			}
		}
		rules.removeAll(remove);
	}

	// if applyExceptions is false then returns both blocking and exception
	// rules,
	// else returns the first matching exception rule
	public List<AdBlockRule> getMatchedRules(String url, boolean applyExceptions)
			throws MalformedURLException {
		List<AdBlockRule> retval = new ArrayList<AdBlockRule>();
		for (AdBlockRule rule : rules) {
			boolean applyrule = false;
			Matcher matcher = Pattern.compile(rule.getRegex()).matcher(url);
			if (!matcher.find()) {
				continue;
			} else {
				applyrule = true;
				for (String opt : rule.getOptions()) {
					if (opt.startsWith("domain=")) {
						applyrule = this.evalDomainOption(url,
								opt.replaceFirst("domain=", ""));
					}
				}
			}
			if (applyrule) {
				retval.add(rule);
			}
		}

		// exception rules negate all other matching rules
		if (applyExceptions) {
			for (AdBlockRule rule : retval) {
				if (rule.isException()) {
					retval.clear();
					retval.add(rule);
					break;
				}
			}
		}
		return retval;
	}

	private boolean evalDomainOption(String url, String optval)
			throws MalformedURLException {
		boolean retval = false;
		String hostname = new URL(url).getHost();
		String[] optdomains = optval.split("\\|");
		for (String optdomain : optdomains) {
			boolean newval = true;
			if (optdomain.startsWith("~")) {
				newval = false;
				optdomain = optdomain.replaceFirst("~", "");
			}
			if (hostname.contains(optdomain)) {
				retval = newval;
			}
		}
		return retval;
	}
}
