/*
* Copyright (C) 2013 Chris Neasbitt
* Author: Chris Neasbitt
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package edu.uga.cs.clickminer.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.Proxy;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLDecoder;
import java.net.URLStreamHandlerFactory;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.NameValuePair;
import org.apache.http.client.utils.URLEncodedUtils;

import uk.ac.shef.wit.simmetrics.similaritymetrics.Levenshtein;

public class MutableURL{

	private URL url;
	private enum PART {PROTOCOL, USERINFO, HOST, PORT, PATH, QUERY, REF};
	
	private static final transient Log log = LogFactory.getLog(MutableURL.class);
	
	public MutableURL(MutableURL url) throws MalformedURLException{
		this.url = new URL(url.toString());
	}
	
	public MutableURL(URL url) throws MalformedURLException {
		this.url = new URL(url.toString());
	}
	
	public MutableURL(String url) throws MalformedURLException{
		this.url = new URL(url);
	}
	
	private Map<MutableURL.PART, String> getParts(){
		Map<MutableURL.PART, String> retval = new HashMap<MutableURL.PART, String>();
		retval.put(PART.PROTOCOL, url.getProtocol());
		retval.put(PART.USERINFO, url.getUserInfo());
		retval.put(PART.HOST, url.getHost());
		retval.put(PART.PORT, url.getPort() == -1 ? null : Integer.toString(url.getPort()));
		retval.put(PART.PATH, url.getPath());
		retval.put(PART.QUERY, url.getQuery());
		retval.put(PART.REF, url.getRef());
		return retval;
	}
	
	private void recreate(Map<MutableURL.PART, String> parts) throws MalformedURLException{
		String urlstr = new String();
		if(parts.get(PART.PROTOCOL) != null){
			urlstr += parts.get(PART.PROTOCOL) + "://";
		}
		if(parts.get(PART.USERINFO) != null){
			urlstr += parts.get(PART.USERINFO) + "@";
		}
		if(parts.get(PART.HOST) != null){
			urlstr += parts.get(PART.HOST);
		}
		if(parts.get(PART.PORT) != null){
			urlstr +=  ":" + parts.get(PART.PORT);
		}
		if(parts.get(PART.PATH) != null){
			urlstr += parts.get(PART.PATH);
		}
		if(parts.get(PART.QUERY) != null){
			urlstr += "?" + parts.get(PART.QUERY);
		}
		if(parts.get(PART.REF) != null){
			urlstr += "#" + parts.get(PART.REF);
		}
		url = new URL(urlstr);
	}
	
	public String getAuthority(){
		return url.getAuthority();
	}
	
	public Object getContent() throws IOException{
		return url.getContent();
	}
	
	@SuppressWarnings("rawtypes")
	public Object getContent(Class[] classes) throws IOException{
		return url.getContent(classes);
	}
	
	public int getDefaultPort(){
		return url.getDefaultPort();
	}
	
	public String getFile(){
		return url.getFile();
	}
	
	public String getHost(){
		return url.getHost();
	}
	
	public String getPath(){
		return url.getPath();
	}
	
	public int getPort(){
		return url.getPort();
	}
	
	public String getProtocol(){
		return url.getProtocol();
	}
	
	public String getQuery(){
		return url.getQuery();
	}
	
	public String getRef(){
		return url.getRef();
	}
	
	public String getUserInfo(){
		return url.getUserInfo();
	}
	
	public int hashCode(){
		return url.hashCode();
	}
	
	public URLConnection openConnection() throws IOException{
		return url.openConnection();
	}
	
	public URLConnection openConnection(Proxy proxy) throws IOException{
		return url.openConnection(proxy);
	}
	
	public InputStream openStream() throws IOException{
		return url.openStream();
	}
	
	public static void setURLStreamHandlerFactory(URLStreamHandlerFactory fac){
		URL.setURLStreamHandlerFactory(fac);
	}
	
	public String toExternalForm(){
		return url.toExternalForm();
	}
	
	public URI toURI() throws URISyntaxException{
		return url.toURI();
	}
	
	public void setProtocol(String protocol) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.PROTOCOL, protocol);
		recreate(parts);
	}
	
	public void setUserInfo(String userinfo) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.USERINFO, userinfo);
		recreate(parts);
	}
	
	public void setHost(String host) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.HOST, host);
		recreate(parts);
	}
	
	public void setPort(Integer port) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		if(port == null){
			parts.put(PART.PORT, null);
		} else {
			parts.put(PART.PORT, port.toString());
		}
		recreate(parts);
	}
	
	public void setPath(String path) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.PATH, path);
		recreate(parts);
	}
	
	public void setQuery(String query) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.QUERY, query);
		recreate(parts);
	}
	
	public void setRef(String ref) throws MalformedURLException{
		Map<MutableURL.PART, String> parts = getParts();
		parts.put(PART.REF, ref);
		recreate(parts);
	}
	
	public void stripQuery(){
		try {
			setQuery(null);
		} catch (MalformedURLException e) {
			if(log.isErrorEnabled()){
				log.error("Removing the query section made url malformed.  " +
						"This shouldn't ever happen.", e);
			}
		}
	}
	
	public void stripRef(){
		try {
			setRef(null);
		} catch (MalformedURLException e) {
			if(log.isErrorEnabled()){
				log.error("Removing the reference section made url malformed.  " +
						"This shouldn't ever happen.", e);
			}
		}
	}
	
	public void stripPort(){
		try {
			setPort(null);
		} catch (MalformedURLException e) {
			if(log.isErrorEnabled()){
				log.error("Removing the port section made url malformed.  " +
						"This shouldn't ever happen.", e);
			}
		}
	}
	
	
	public void normalizePath() throws MalformedURLException, URISyntaxException{
		String[] parts = getPath().split(";", 2);
		if(parts.length > 0){
			String newpath;
			String pathpart = parts[0];
			if(pathpart.endsWith("/")){
				pathpart = pathpart.substring(0, pathpart.length() - 1);
			}
			newpath = pathpart;
			if(parts.length > 1){
				newpath += ";" + parts[1];
			}
			setPath(newpath);
		}
		url = url.toURI().normalize().toURL();
	}
	
	private void normalizeEncoding() throws MalformedURLException, UnsupportedEncodingException{
		url = new URL(URLDecoder.decode(url.toString(), "UTF-8"));
	}
	
	public void normalizeUrl() throws MalformedURLException, URISyntaxException, UnsupportedEncodingException {
		normalizePath();
		normalizeEncoding();
	}
	
	public float getSimilarity(MutableURL url){
		return (new Levenshtein().getSimilarity(url.toString(), this.url.toString()));
	}
	
	public float getSimilarity(String url){
		return (new Levenshtein().getSimilarity(url, this.url.toString()));
	}
	
	public boolean approxEquals(MutableURL url){
		return approxEquals(url.toString());
	}
	
	public boolean approxEquals(String url){
		return getSimilarity(url) > 0.95;
	}
	
	public Set<NameValuePair> getQueryParams()
			throws MalformedURLException, URISyntaxException {
		return new HashSet<NameValuePair>(URLEncodedUtils.parse(toURI(), "UTF-8"));
	}
	
	private void equalsNormalizer(MutableURL url) throws MalformedURLException, UnsupportedEncodingException, URISyntaxException{
		url.normalizeUrl();
		url.stripRef();
		url.stripQuery();
		
		if(url.getPort() == url.getDefaultPort()){
			url.stripPort();
		}
	}
	
	@Override
	public boolean equals(Object obj) {
		try{
			if(obj instanceof MutableURL){
				MutableURL other = new MutableURL((MutableURL)obj);
				MutableURL cur = new MutableURL(this);
				
				Set<NameValuePair> otherParams = other.getQueryParams();
				Set<NameValuePair> curParams = cur.getQueryParams();
				
				equalsNormalizer(other);
				equalsNormalizer(cur);
								
				return other.toString().equals(cur.toString()) && otherParams.equals(curParams);
			}
		} catch (URISyntaxException e) {
			if(log.isDebugEnabled()){
				log.debug("", e);
			}
		} catch (Exception e) {
			if(log.isErrorEnabled()){
				log.error("", e);
			}
		}
		return false;
	}
	
	
	@Override
	public String toString(){
		return url.toString();
	}
	
}
